#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <commctrl.h>
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "comctl32.lib")
#pragma comment(lib, "gdi32.lib")

const struct s2s_t {
    const char *name;
    int val;
    const char *help;
} s2s[] = {
    {"SW_FORCEMINIMIZE", SW_FORCEMINIMIZE, "Windows 2000/XP: Minimizes a window, even if the thread that owns the window is hung. This flag should only be used when minimizing windows from a different thread."},
    {"SW_HIDE", SW_HIDE, "Hides the window and activates another window."},
    {"SW_MAXIMIZE", SW_MAXIMIZE, "Maximizes the specified window."},
    {"SW_MINIMIZE", SW_MINIMIZE, "Minimizes the specified window and activates the next top-level window in the Z order."},
    {"SW_RESTORE", SW_RESTORE, "Activates and displays the window. If the window is minimized or maximized, the system restores it to its original size and position. An application should specify this flag when restoring a minimized window."},
    {"SW_SHOW", SW_SHOW, "Activates the window and displays it in its current size and position."},
    {"SW_SHOWDEFAULT", SW_SHOWDEFAULT, "Sets the show state based on the SW_ value specified in the STARTUPINFO structure passed to the CreateProcess function by the program that started the application."},
    {"SW_SHOWMAXIMIZED", SW_SHOWMAXIMIZED, "Activates the window and displays it as a maximized window."},
    {"SW_SHOWMINIMIZED", SW_SHOWMINIMIZED, "Activates the window and displays it as a minimized window."},
    {"SW_SHOWMINNOACTIVE", SW_SHOWMINNOACTIVE, "Displays the window as a minimized window.\r\nThis value is similar to SW_SHOWMINIMIZED, except the window is not activated."},
    {"SW_SHOWNA", SW_SHOWNA, "Displays the window in its current size and position.\r\nThis value is similar to SW_SHOW, except the window is not activated."},
    {"SW_SHOWNOACTIVATE", SW_SHOWNOACTIVATE, "Displays a window in its most recent size and position.\r\nThis value is similar to SW_SHOWNORMAL, except the window is not actived."},
    {"SW_SHOWNORMAL", SW_SHOWNORMAL, "Activates and displays a window. If the window is minimized or maximized, the system restores it to its original size and position. An application should specify this flag when displaying the window for the first time."},
    {0, 0}
};

extern HWND ListW;
BOOL CALLBACK EnumWindowsProc(HWND hwnd, LPARAM lParam) {
	char s[1024]; char *c = (char*)lParam;
	GetWindowText(hwnd, s, 1023);
    //CharToOem(s, s);
    if (c? (strstr(s, c) != 0) : *s != 0) {
        //printf("0x%08X: '%s'\n", hwnd, s);
        LVITEM item;
        item.mask = LVIF_TEXT, item.iItem = 0, item.iSubItem = 0;
        item.stateMask = 0, item.lParam = 0;
        item.pszText = s;
        int I = ListView_InsertItem(ListW, &item);
        if (I >= 0) {
            char tmp[32]; sprintf(tmp, "0x%08X", hwnd);
            item.pszText = tmp;
            item.iItem=I;
            item.iSubItem=1;
            ListView_SetItem(ListW, &item);
        }
    }
	return true;
}

const int NumButtonsInRow = 1; // >0
LRESULT CALLBACK WindowProc (HWND,UINT,WPARAM,LPARAM);
HINSTANCE hInst;
HWND MainWindow, hwndToolTip, ListW,  EditHandleW, EditMaskW, Buttonqae, Buttonqst;
int ButtonNum;
HWND *hMWButtons = 0;
const char *myname = "ShowWindowClass";
HFONT DlgThinFnt;
int butt_szxs[NumButtonsInRow];
int butt_stxs[NumButtonsInRow];
int ListWWidth = 300, ListWHeight = 100;
int MainWindowWidth = ListWWidth + 10 + 320, MainWindowHeight = 200;

int WINAPI WinMain (HINSTANCE hTI, HINSTANCE hPI, LPSTR lpszargs, int nWinMode) {
    HWND hwnd;
    MSG msg;
    WNDCLASS wcl;

    hInst = hTI;
    
    HWND w2 = FindWindow(myname, NULL);
	if (w2) { ShowWindow(w2, SW_RESTORE); return 0; }
	
	InitCommonControls();

    memset(&wcl, 0, sizeof wcl);
    wcl.hInstance = hTI, wcl.lpszClassName = myname;
    wcl.lpfnWndProc = WindowProc, wcl.style=CS_HREDRAW | CS_VREDRAW;
    wcl.hIcon = 0;//LoadIcon (hTI,"someicon");
    wcl.hCursor = LoadCursor (NULL, IDC_ARROW);
    wcl.lpszMenuName = 0;//"SomeMenu";
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hbrBackground = (HBRUSH)GetStockObject (LTGRAY_BRUSH);

    if (!RegisterClass(&wcl)) return 0;

	//LoadSettings();
    for(ButtonNum = 0; s2s[ButtonNum].name; ButtonNum++);
    MainWindowHeight = 120 + 21 * ((ButtonNum + NumButtonsInRow - 1) / NumButtonsInRow);
    butt_szxs[0] = butt_stxs[0] = 0;
    hMWButtons = new HWND[ButtonNum];
    memset(hMWButtons, 0, ButtonNum * sizeof(*hMWButtons));
    hwnd = CreateWindow(myname,
      "ShowWindow", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT,
      MainWindowWidth, //Width
      MainWindowHeight, //Height
      HWND_DESKTOP, NULL, hTI, NULL
    );
	MainWindow = hwnd;
    ShowWindow(hwnd, nWinMode);

    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return msg.wParam;
}

struct fchar {
  char *c;
  fchar() : c(0) {}
  explicit fchar(void *v) : c((char*)v) {}
  fchar(fchar &f) : c(f.c) { f.c = 0; }
  operator char*&() { return c;}
  operator unsigned char*&() { return (unsigned char*&)c;}
  char &operator *() { return *c; }
  operator const char*() const { return c;}
  operator const unsigned char*() const { return (unsigned char*)c;}
  //    operator bool() const { return c != 0; }
  ~fchar() { free(c); }
  UINT GetWindowText(HWND hwnd);
private:
//  fchar(const fchar &f);
};
UINT fchar::GetWindowText(HWND hwnd) { //single-threaded!
  DWORD L = SendMessage(hwnd, WM_GETTEXTLENGTH,0,0);
  c = (char*)realloc(c, L + 1);
  return ::GetWindowText(hwnd, c, L + 1);
}

bool qae = false, qst = false;
fchar qst_nf_sav;

void RefreshWindowsList() {
    ListView_DeleteAllItems(ListW);
    fchar c; c.GetWindowText(EditMaskW);
    EnumWindows(EnumWindowsProc, (LPARAM)(c.c && *c.c? c.c : qae? "" : 0));
}

LRESULT CALLBACK WindowProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    HDC hdc;
    TOOLINFO ti;
    LVCOLUMN lvcol;
//    PAINTSTRUCT ps;
//    RECT *lpR;
    int n,k;
//    HGDIOBJ ltgbr, lasthhh;
//    lpR=lpR; //for Warning

    switch (msg) {
    case WM_CREATE:
        DlgThinFnt=CreateFont(15,5,0,0,FW_LIGHT,FALSE,FALSE,FALSE,DEFAULT_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY,DEFAULT_PITCH | FF_DONTCARE, "Arial Cyr");
        hwndToolTip = CreateWindowEx( 0, TOOLTIPS_CLASS, NULL, WS_POPUP /*| TTS_ALWAYSTIP*/,
            CW_USEDEFAULT, CW_USEDEFAULT, 20, 20, hwnd, NULL, hInst, NULL);
        if (!hwndToolTip) MessageBeep(MB_OK);
        SendMessage(hwndToolTip, TTM_SETMAXTIPWIDTH, 0, 300);
        hdc = GetDC(hwnd);
        for(n = 0; s2s[n].name; n++) {
            SIZE sz;
            if (!GetTextExtentPoint32(hdc, s2s[n].name, strlen(s2s[n].name), &sz)) continue;
            if (butt_szxs[n % NumButtonsInRow] + 4 < sz.cx) butt_szxs[n % NumButtonsInRow] = sz.cx + 4;
        }
        ReleaseDC(hwnd, hdc);
        for(n = 1; n < NumButtonsInRow; n++) butt_stxs[n] = butt_stxs[n - 1] + butt_szxs[n - 1] + 3;
        MainWindowWidth = ListWWidth + 10 + butt_stxs[NumButtonsInRow - 1] + butt_szxs[NumButtonsInRow - 1] + 3;
        ZeroMemory(&ti, sizeof(ti));
        ti.cbSize = sizeof(ti);
        ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
        ti.hwnd = hwnd;
        ti.hinst = hInst;
        ti.lpszText = LPSTR_TEXTCALLBACK;
        for(n=0; s2s[n].name; n++) {
            hMWButtons[n] = CreateWindowEx(0,"BUTTON", s2s[n].name,
                WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                ListWWidth + 2 + butt_stxs[n % NumButtonsInRow], 1 + 21 * (n / NumButtonsInRow),
                butt_szxs[n % NumButtonsInRow], 20, hwnd, (HMENU)(200 + n), hInst, NULL);
            SendMessage(hMWButtons[n], WM_SETFONT, (WPARAM)DlgThinFnt, MAKELPARAM(FALSE,0));
            ti.uId = (UINT)hMWButtons[n];
            SendMessage(hwndToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
        }
        EditHandleW = CreateWindow("EDIT", "", WS_CHILD | WS_VISIBLE | WS_TABSTOP, 0, 0, ListWWidth, 20, hwnd, (HMENU)98, hInst, 0);
        ti.uId = (UINT)EditHandleW;
        SendMessage(hwndToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
        EditMaskW =  CreateWindow("EDIT", "", WS_CHILD | WS_VISIBLE | WS_TABSTOP, 0, 22, ListWWidth - 40, 20, hwnd, (HMENU)97, hInst, 0);
        ti.uId = (UINT)EditMaskW;
        SendMessage(hwndToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
        Buttonqae = CreateWindowEx(0, "BUTTON", "E", WS_CHILD | WS_VISIBLE | BS_PUSHLIKE | BS_AUTOCHECKBOX,
                ListWWidth - 20, 22, 20, 20, hwnd, (HMENU)96, hInst, NULL);
        ti.uId = (UINT)Buttonqae;
        SendMessage(hwndToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
        Buttonqst = CreateWindowEx(0, "BUTTON", "T", WS_CHILD | WS_VISIBLE | BS_PUSHLIKE | BS_AUTOCHECKBOX,
                ListWWidth - 40, 22, 20, 20, hwnd, (HMENU)95, hInst, NULL);
        ti.uId = (UINT)Buttonqst;
        SendMessage(hwndToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
        ListW = CreateWindowEx(WS_EX_CLIENTEDGE, WC_LISTVIEW, "none",
            WS_CHILD | WS_VISIBLE | LVS_REPORT | WS_TABSTOP |
            LVS_SORTASCENDING | LVS_SINGLESEL | LVS_SHOWSELALWAYS,
            0, 44, ListWWidth, ListWHeight, hwnd, (HMENU)99, hInst, NULL);
        lvcol.mask = LVCF_FMT | LVCF_SUBITEM | LVCF_TEXT | LVCF_WIDTH, lvcol.fmt = LVCFMT_LEFT;
        lvcol.iSubItem = 0, lvcol.pszText = "Name", lvcol.cx = ListWWidth / 2;
        ListView_InsertColumn(ListW,0,&lvcol);
        lvcol.iSubItem = 1, lvcol.pszText = "Handle", lvcol.cx = ListWWidth / 2 - 20;
        ListView_InsertColumn(ListW,1,&lvcol);
        ti.uId = (UINT)ListW;
        SendMessage(hwndToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
        RefreshWindowsList();
        SetWindowPos(hwnd, 0, 0, 0, MainWindowWidth, MainWindowHeight, SWP_NOMOVE | SWP_NOZORDER);
        break;
    
    case WM_SIZE:
        if (wParam == SIZE_MINIMIZED) break;
        MainWindowWidth = LOWORD(lParam);
        MainWindowHeight = HIWORD(lParam);
        ListWWidth = MainWindowWidth - (10 + butt_stxs[NumButtonsInRow - 1] + butt_szxs[NumButtonsInRow - 1] + 3);
        ListWHeight = MainWindowHeight - 44;
        if (ListWWidth < 40) ListWWidth = 40;
        if (ListWHeight < 40) ListWHeight = 40;
        SetWindowPos(ListW, 0, 44, 0, ListWWidth, ListWHeight, SWP_NOMOVE | SWP_NOZORDER);
        SetWindowPos(EditHandleW, 0, 0, 0, ListWWidth, 20, SWP_NOMOVE | SWP_NOZORDER);
        SetWindowPos(EditMaskW, 0, 22, 0, ListWWidth - 40, 20, SWP_NOMOVE | SWP_NOZORDER);
        SetWindowPos(Buttonqae, 0, ListWWidth - 20, 22, 20, 20, SWP_NOZORDER);
        SetWindowPos(Buttonqst, 0, ListWWidth - 40, 22, 20, 20, SWP_NOZORDER);
        for(n = 0; n < ButtonNum; n++) {
            SetWindowPos(hMWButtons[n], 0, ListWWidth + 2 + butt_stxs[n % NumButtonsInRow], 1 + 21 * (n / NumButtonsInRow),
                butt_szxs[n % NumButtonsInRow], 20, SWP_NOZORDER);
        }
        break;

    case WM_COMMAND:
        if (LOWORD(wParam) >= 200 && LOWORD(wParam) < 200 + ButtonNum) {
            fchar es; es.GetWindowText(EditHandleW);
            HWND hwnd1 = (HWND)strtoul(es, 0, 0);
            if (!hwnd1) break;
            char buf[128];
            int val = s2s[LOWORD(wParam) - 200].val;
            sprintf(buf, "ShowWindow(0x%08X, %i)...", hwnd1, val);
            SetWindowText(hwnd, buf);
            int rv = ShowWindow(hwnd1, val);
            sprintf(buf, "ShowWindow(0x%08X, %i) returned %i (0x%08X)\n", hwnd1, val, rv, rv);
            SetWindowText(hwnd, buf);
        }
        switch (LOWORD(wParam)) {
        case 95:
            qst = SendMessage(Buttonqst, BM_GETCHECK, 0, 0) != 0;
            EnableWindow(ListW, !qst);
            SetWindowText(Buttonqae, qst? "A" : "E");
            if (qst) {
                qst_nf_sav.GetWindowText(EditMaskW);
                SendMessage(Buttonqae, BM_SETCHECK, 0, 0);
                SendMessage(Buttonqae, BM_SETSTYLE, BS_PUSHBUTTON, 0);
                fchar es, ns; es.GetWindowText(EditHandleW);
                HWND hwnd1 = (HWND)strtoul(es, 0, 0);
                if (hwnd1) {
                    ns.GetWindowText(hwnd1);
                    SetWindowText(EditMaskW, ns);
                }
            } else {
                SetWindowText(EditMaskW, qst_nf_sav);
                SendMessage(Buttonqae, BM_SETSTYLE, BS_PUSHLIKE | BS_AUTOCHECKBOX, 0);
                SendMessage(Buttonqae, BM_SETCHECK, qae, 0);
            }
            break;
        case 96:
            if (!qst) {
                qae = SendMessage(Buttonqae, BM_GETCHECK, 0, 0) != 0;
                RefreshWindowsList();
            } else {
                fchar es, ns; es.GetWindowText(EditHandleW);
                ns.GetWindowText(EditMaskW);
                HWND hwnd1 = (HWND)strtoul(es, 0, 0);
                if (!hwnd1) break;
                if (!qae && (!ns.c || !*ns.c)) {
                    MessageBox(hwnd, "You didn't allow empty window titles", "OPS!", MB_OK);
                    break;
                }
                SetWindowText(hwnd1, ns);
            }
            break;
        case 97:
            if (!qst && HIWORD(wParam)==EN_UPDATE) RefreshWindowsList();
            break;
        }
        break;

    case WM_NOTIFY:
        k = ((LPNMHDR)lParam)->idFrom;
        switch (((LPNMHDR)lParam)->code) {
            char buf[32];
        case TTN_NEEDTEXT:
            if (k == (int)EditHandleW) {
                ((LPTOOLTIPTEXT)lParam)->lpszText = "Place for window handle";
                return 1;
            }
            if (k == (int)EditMaskW) {
                ((LPTOOLTIPTEXT)lParam)->lpszText = qst?
                    "New window title" : "Window names listed must have this substring";
                return 1;
            }
            if (k == (int)ListW) {
                ((LPTOOLTIPTEXT)lParam)->lpszText = "List of top-level windows\r\nRight-click to refresh";
                return 1;
            }
            if (k == (int)Buttonqae) {
                ((LPTOOLTIPTEXT)lParam)->lpszText = qst? "Apply" : "Select to show windows with empty titles";
                return 1;
            }
            if (k == (int)Buttonqst) {
                ((LPTOOLTIPTEXT)lParam)->lpszText = "Select to set window text";
                return 1;
            }
            for(n = 0; n < ButtonNum && k != (int)hMWButtons[n]; n++);
            if (n >= ButtonNum) break;
            ((LPTOOLTIPTEXT)lParam)->lpszText = (char*)s2s[n].help;
            return 1;
        case LVN_ITEMCHANGED:
            n = ListView_GetNextItem(ListW, -1, LVNI_SELECTED);
            if (n < 0) break;
            ListView_GetItemText(ListW, n, 1, buf, 32);
            SetWindowText(EditHandleW, buf);
            break;
	    case NM_RCLICK:
            if (k == 99) RefreshWindowsList();
            break;
        }
        break;

    case WM_CLOSE:
        DestroyWindow(hwnd);
        break;
        
    case WM_DESTROY:
        DeleteObject(DlgThinFnt);
        PostQuitMessage(0);
        break;
    default:
        return DefWindowProc (hwnd, msg, wParam, lParam);

    }
    return 0;
}
